/**
 * jQuery Yii GridView plugin file extension.
 *
 * This extension overwrites the $.fn.yiiGridView namespace with an alternative
 * constructor function and adds some helper functions, in order to allow the GridView
 * to create a hash history.
 *
 * @author Rupert, Red Rabbit Studio
 * @link http://www.rabbitstyle.com/
 * @copyright Copyright &copy; 2010-2011 Red Rabbit Studio
 * @license http://www.opensource.org/licenses/mit-license.php MIT License
 * @version $Id: jquery.rr.yiigridview.js 5 2011-05-24
 */

(function($) {
	/* backup all the $.fn.yiiGridView objects so that we can reinsert them after overriding
	   the constructor function */
	var tmpGridView = {};
	for (attr in $.fn.yiiGridView) {
		tmpGridView[attr] = $.fn.yiiGridView[attr];
	}
	
	/**
	 * yiiGridView set function.
	 * @param options map settings for the grid view. Available options are as follows:
	 * - ajaxUpdate: array, IDs of the containers whose content may be updated by ajax response
	 * - ajaxVar: string, the name of the GET variable indicating the ID of the element triggering the AJAX request
	 * - pagerClass: string, the CSS class for the pager container
	 * - tableClass: string, the CSS class for the table
	 * - selectableRows: integer, the number of rows that can be selected
	 * - updateSelector: string, the selector for choosing which elements can trigger ajax requests
	 * - hashHistory: boolean, whether ajax calls should be reflected in hash history
	 * - beforeAjaxUpdate: function, the function to be called before ajax request is sent
	 * - afterAjaxUpdate: function, the function to be called after ajax response is received
	 * - ajaxUpdateError: function, the function to be called if an ajax error occurs
	 * - selectionChanged: function, the function to be called after the row selection is changed
	 */
	$.fn.yiiGridView = function(options) {
		return this.each(function(){
			var settings = $.extend({}, $.fn.yiiGridView.defaults, options || {});
			var $this = $(this);
			var id = $this.attr('id');
			if(settings.updateSelector === undefined) {
				settings.updateSelector = '#'+id+' .'+settings.pagerClass.replace(/\s+/g,'.')+' a, #'+id+' .'+settings.tableClass.replace(/\s+/g,'.')+' thead th a';
			}
			
			$.fn.yiiGridView.settings[id] = settings;

			/* ********* */
			// initialise the GET parameter array
			var getParams = [];
			// if hashHistory is enabled, define the default selectors for paging and sorting links
			if (settings.hashHistory) {
				if (settings.pageUpdater === undefined) {
					settings.pageUpdater = '#' + id + ' .' + settings.pagerClass.replace(/\s+/g,'.') + ' a';
				}
				if (settings.sortUpdater === undefined) {
					settings.sortUpdater = '#' + id + ' .' + settings.tableClass.replace(/\s+/g,'.') + ' thead th a';
				}
			}
			/* --------- */
			
			if(settings.ajaxUpdate.length > 0) {
				/* ********* */
				// if hashHistory is enabled bind the appropriate click functions, otherwise continue with default
				if (settings.hashHistory) {
					// only use pager links if pagination is defined
					if (settings.pageUpdater !== '' && settings.pageVar !== undefined) {
						// add the pagination parameter to the GET parameters array
						getParams.push(settings.pageVar);
						$(settings.pageUpdater).die('click').live('click',function(){
							// get the query parameter objects from the link location
							var linkParams = $.deparam.querystring($(this).attr('href'));
							// create a params object, settings the pageVar param to
							// null by default, and overriding it if it is in the query parameters
							var params = {};
							params[settings.pageVar] = null;
							if (linkParams[settings.pageVar] !== undefined) {
								params[settings.pageVar] = linkParams[settings.pageVar];
							}
							// update the current hash with the new pageVar state
							$.fn.yiiGridView.updateHash(params);
							return false;
						});
					}
					// only use sorter links if sorting is defined
					if (settings.sortUpdater !== '' && settings.sortVar !== undefined) {
						// add the sort parameter to the GET parameters array
						getParams.push(settings.sortVar);
						$(settings.sortUpdater).die('click').live('click',function(){
							// get the query parameter objects from the link location
							var linkParams = $.deparam.querystring($(this).attr('href'));
							// create a params object, settings the sortVar param to
							// null by default, and overriding it if it is in the query parameters
							var params = {};
							params[settings.sortVar] = null;
							if (linkParams[settings.sortVar] !== undefined) {
								params[settings.sortVar] = linkParams[settings.sortVar];
							}
							// update the current hash with the new pageVar state
							$.fn.yiiGridView.updateHash(params);
							return false;
						});
					}
				} else {
					/* --------- */
					$(settings.updateSelector).die('click').live('click',function(){
						$.fn.yiiGridView.update(id, {url: $(this).attr('href')});
						return false;
					});
					/* ********* */
				}
				/* --------- */
			}

			var inputSelector='#'+id+' .'+settings.filterClass+' input, '+'#'+id+' .'+settings.filterClass+' select';
			/* ********* */
			if (settings.hashHistory) {
				// if hash history is enabled and inputUpdater undefined, get the default selector
				if (settings.inputUpdater === undefined) {
					settings.inputUpdater = inputSelector;
				}
				if (settings.inputUpdater !== '') {
					// loop through all the inputs, adding each one to the GET parameter array
					$(settings.inputUpdater).each(function() {
						getParams.push($(this).attr('name'));
					});
					$('body').undelegate(settings.inputUpdater, 'change').delegate(settings.inputUpdater, 'change', function(){
						var context = $(this);
						var params = {};
						// artificially create the param object from the input name and value
						// (allows variable depth objects like Model[attribute] = value
						param = $.deparam(context.attr('name') + '=' + context.val());
						// if the input value is empty, then we need to set the lowest-child
						// attribute to null, so that that part of the object is removed from
						// the state parameters
						if (context.val() === '') {
							$.fn.yiiGridView.setNull(param);
						}
						// loop through each attribute of the param object in order to insert
						// the param object into the params object
						for(paramObj in param) {
							params[paramObj] = param[paramObj];
						}
						// if pagination is enabled, then by default we reset the page to 1 on filtering
						if (settings.pageVar !== undefined && settings.pageUpdater !== '') {
							params[settings.pageVar] = '1';
						}
						// update the current hash with the new input state
						$.fn.yiiGridView.updateHash(params);
						return false;
					});
				}
				/* ********* */
			} else {
				/* --------- */
				$('body').undelegate(inputSelector, 'change').delegate(inputSelector, 'change', function(){
					var data = $(inputSelector).serialize();
					if(settings.pageVar!==undefined) {
						data += '&'+settings.pageVar+'=1';
					}
					$.fn.yiiGridView.update(id, {data: data});
				});
			}
			/* --------- */

			$.fn.yiiGridView.selectCheckedRows(id);

			if(settings.selectableRows > 0) {
				$('#'+id+' .'+settings.tableClass+' > tbody > tr').die('click').live('click',function(e){
					if('checkbox' !== e.target.type){
						if(settings.selectableRows === 1) {
							$(this).siblings().removeClass('selected');
						}

						var isRowSelected=$(this).toggleClass('selected').hasClass('selected');
						$('input.select-on-check',this).each(function(){
							if(settings.selectableRows === 1) {
								$("input[name='"+this.name+"']").attr('checked',false);
							}
							this.checked=isRowSelected;
							var sboxallname=this.name.substring(0,this.name.length-2)+'_all';	//.. remove '[]' and add '_all'
							$("input[name='"+sboxallname+"']").attr('checked', $("input[name='"+this.name+"']").length===$("input[name='"+this.name+"']:checked").length);
						});
						if(settings.selectionChanged !== undefined) {
							settings.selectionChanged(id);
						}
					}
				});
			}

			$('#'+id+' .'+settings.tableClass+' > tbody > tr > td > input.select-on-check').die('click').live('click',function(){
					if(settings.selectableRows === 0) {
						return false;
					}

					var $row=$(this).parent().parent();
					if(settings.selectableRows === 1){
						$row.siblings().removeClass('selected');
						$("input:not(#"+this.id+")[name='"+this.name+"']").attr('checked',false);
					}
					else {
						$('#'+id+' .'+settings.tableClass+' > thead > tr > th >input.select-on-check-all').attr('checked', $("input.select-on-check").length===$("input.select-on-check:checked").length);
					}

					$row.toggleClass('selected',this.checked);
					if(settings.selectionChanged !== undefined) {
						settings.selectionChanged(id);
					}
					return true;
			});

			if(settings.selectableRows > 1) {
				$('#'+id+' .'+settings.tableClass+' > thead > tr > th > input.select-on-check-all').die('click').live('click',function(){
					var checkedall=this.checked;
					var name=this.name.substring(0,this.name.length-4)+'[]';	//.. remove '_all' and add '[]'
					$("input[name='"+name+"']").each(function() {
						this.checked=checkedall;
						$(this).parent().parent().toggleClass('selected',checkedall);
					});
					if(settings.selectionChanged !== undefined) {
						settings.selectionChanged(id);
					}
				});
			}
			/* ********* */
			// if hash history is enabled and there are GET parameters to be monitored...
			if (settings.hashHistory && getParams !== undefined && getParams.length > 0) {
				var x = 0;
				var group = id + 'GridView';
				// register the hash handler group for this gridview, using the first GET parameter
				$.registerHashHandler({
					parameter: getParams[x]
				}, getParams[x], group, {
					// the callback for the hash handler that allows us to update the gridview
					// according to the current hash state
					callback: function(params, error) {
						// merge the current location and hash fragment parameters
						var pageParams = $.extend($.fn.yiiGridView.parseParams($.param.querystring()), $.fn.yiiGridView.parseParams($.param.fragment()));
						// loop through the changed parameters
						for(param in params) {
							if (params[param].next === null) {
								if (pageParams[param] !== undefined) {
									delete pageParams[param];
								}
							} else {
								// otherwise set it to the new value
								pageParams[param] = params[param].next;
							}
						}
						/*
						 * make the Ajax update, using the new Url, and overriding
						 * the error function to include the hash error callback
						 * (so that the hash can be changed back to its previous
						 * state in case of error)
						 */
						var data = '';
						// serialize the parameters
						for(param in pageParams) {
							data += '&' + param + '=' + pageParams[param];
						}
						// make the ajax update request
						$.fn.yiiGridView.update(id, {
							url: window.location.href.split('?')[0],
							data: data,
							error: function(XMLHttpRequest, textStatus, errorThrown) {
								$('#'+id).removeClass(settings.loadingClass);
								alert(XMLHttpRequest.responseText);
								error();
							},
							cache: false
						});
					}
				});
				// loop through the remaining GET parameters, adding each one to the hash handler
				var y = getParams.length;
				for(x = 1; x < y; x++) {
					$.registerHashHandler({
						parameter: getParams[x]
					}, getParams[x], group, {});
				}
			}
			// trigger the hash change event for the first time in case an ajax link has been accessed
			$(window).trigger('hashchange');
			/* --------- */
		});
	};
	
	/* ********* */
	/**
	 * Updates the hash state with the parameter objects in params
	 * @param params the object containing the parameter objects to be updated
	 */
	$.fn.yiiGridView.updateHash = function(params) {
		// get the parameters from the location
		var pageParams = $.deparam.querystring();
		// get the current hash state
		var state = $.bbq.getState();
		// merge the params with the current hash state
		state = $.extend(true, state, params);
		// check the updated params for null values or redundant values that can be left out of the hash state
		$.fn.yiiGridView.checkCompare(params, pageParams, state);
		// update the hash state
		$.bbq.pushState(state, 2);
	};

	/**
	 * Parses a querystring to return an array of parameters
	 * @param querystring the querystring to be parsed
	 * @return object an associative array of parameter names and values
	 */
	$.fn.yiiGridView.parseParams = function (querystring) {
	    var queryString = {};
	    decodeURI(querystring).replace(
	        new RegExp("([^?=&]+)(=([^&]*))?", "g"),
	        function($0, $1, $2, $3) { queryString[$1] = $3; }
	    );
	    return queryString;
	};
	
	/**
	 * Recursively sets the lowest-level child attribute of obj to null
	 * Used for objects generated from an Html form input tag, which
	 * means that each child attribute may only have one sub-attribute.
	 * @param obj the object whose lowest-level child is to be set to null
	 */
	$.fn.yiiGridView.setNull = function(obj) {
		for (attr in obj) {
			if (typeof(obj[attr]) === 'object') {
				$.fn.yiiGridView.setNull(obj[attr]);
			}
			else {
				obj[attr] = null;
			}
		}
	};
	
	/**
	 * Recursively compares a new params object, the parameters included in
	 * the current location, and the current hash state, to determine whether
	 * any parameters should be deleted from the hash state (such as those
	 * whose values are null, or whose values are the same as those given by
	 * the location Url)
	 * @param contained the params object containing the parameter objects
	 *                  that have been updated
	 * @param container the location parameters object containing the
	 *                  parameter objects derived from the location Url
	 * @param state the current state object containing the parameter objects
	 *              that may be updated.
	 */
	$.fn.yiiGridView.checkCompare = function (contained, container, state) {
		for (attr in contained) {
			if (typeof(contained[attr]) === 'object') {
				var subContainer = null;
				if (container !== null && container[attr] !== undefined) {
					subContainer = container[attr];
				}
				if ($.fn.yiiGridView.checkCompare(contained[attr], subContainer, state[attr])) {
					delete state[attr];
				}
			}
			else {
				if ((contained[attr] === null && (container === null || container[attr] === undefined)) || (container !== null && container[attr] !== undefined && contained[attr] === container[attr])) {
					delete state[attr];
				}
			}
		}
		var count = 0;
		for (attr in state) {
			count++;
		}
		if (count === 0) {
			return true;
		}
		return false;
	};
	
	$.fn.yiiGridView.defaults = {
		ajaxUpdate: [],
		ajaxVar: 'ajax',
		pagerClass: 'pager',
		loadingClass: 'loading',
		filterClass: 'filters',
		tableClass: 'items',
		selectableRows: 1,
		hashHistory: false
		// updateSelector: '#id .pager a, '#id .grid thead th a',
		// beforeAjaxUpdate: function(id) {},
		// afterAjaxUpdate: function(id, data) {},
		// selectionChanged: function(id) {},
	};
	
	/* reinsert the $.fn.yiiGridView objects that we backup up before
	   overwriting the constructor function */
	for(attr in tmpGridView) {
		// only reinsert the object if we haven't created our own
		if ($.fn.yiiGridView[attr] === undefined) {
			$.fn.yiiGridView[attr] = tmpGridView[attr];
		}
	}
}(jQuery));