/**
 * jQuery Yii GridView plugin file extension.
 *
 * This extension overwrites the $.fn.yiiGridView namespace with an alternative
 * constructor function and adds some helper functions, in order to allow the GridView
 * to create a hash history.
 *
 * @author Rupert, Red Rabbit Studio
 * @link http://www.rabbitstyle.com/
 * @copyright Copyright &copy; 2010-2011 Red Rabbit Studio
 * @license http://www.opensource.org/licenses/mit-license.php MIT License
 * @version $Id: jquery.rr.yiigridview.js 5 2011-05-24
 */

(function($) {
	/* backup all the $.fn.yiiListView objects so that we can reinsert them after overriding
	   the constructor function */
	var tmpListView = {};
	for (attr in $.fn.yiiListView) {
		tmpListView[attr] = $.fn.yiiListView[attr];
	}
	
	/**
	 * yiiListView set function.
	 * @param options map settings for the list view. Available options are as follows:
	 * - ajaxUpdate: array, IDs of the containers whose content may be updated by ajax response
	 * - ajaxVar: string, the name of the GET variable indicating the ID of the element triggering the AJAX request
	 * - pagerClass: string, the CSS class for the pager container
	 * - sorterClass: string, the CSS class for the sorter container
	 * - updateSelector: string, the selector for choosing which elements can trigger ajax requests
	 * - hashHistory: boolean, whether ajax calls should be reflected in hash history
	 * - beforeAjaxUpdate: function, the function to be called before ajax request is sent
	 * - afterAjaxUpdate: function, the function to be called after ajax response is received
	 */
	$.fn.yiiListView = function(options) {
		return this.each(function(){
			var settings = $.extend({}, $.fn.yiiListView.defaults, options || {});
			var $this = $(this);
			var id = $this.attr('id');
			if(settings.updateSelector == undefined) {
				settings.updateSelector = '#'+id+' .'+settings.pagerClass.replace(/\s+/g,'.')+' a, #'+id+' .'+settings.sorterClass.replace(/\s+/g,'.')+' a';
			}
			$.fn.yiiListView.settings[id] = settings;
			
			/* ********* */
			// initialise the GET parameter array
			var getParams = [];
			// if hashHistory is enabled, define the default selectors for paging and sorting links
			if (settings.hashHistory) {
				if (settings.pageUpdater === undefined) {
					settings.pageUpdater = '#' + id + ' .' + settings.pagerClass.replace(/\s+/g,'.') + ' a';
				}
				if (settings.sortUpdater === undefined) {
					settings.sortUpdater = '#' + id + ' .' + settings.sorterClass.replace(/\s+/g,'.') + ' a';
				}
			}
			/* --------- */

			if(settings.ajaxUpdate.length > 0) {
				/* ********* */
				// if hashHistory is enabled bind the appropriate click functions, otherwise continue with default
				if (settings.hashHistory) {
					// only use pager links if pagination is defined
					if (settings.pageUpdater !== '' && settings.pageVar !== undefined) {
						// add the pagination parameter to the GET parameters array
						getParams.push(settings.pageVar);
						$(settings.pageUpdater).die('click').live('click',function(){
							// get the query parameter objects from the link location
							var linkParams = $.deparam.querystring($(this).attr('href'));
							// create a params object, settings the pageVar param to
							// null by default, and overriding it if it is in the query parameters
							var params = {};
							params[settings.pageVar] = null;
							if (linkParams[settings.pageVar] !== undefined) {
								params[settings.pageVar] = linkParams[settings.pageVar];
							}
							// update the current hash with the new pageVar state
							$.fn.yiiListView.updateHash(params);
							return false;
						});
					}
					// only use sorter links if sorting is defined
					if (settings.sortUpdater !== '' && settings.sortVar !== undefined) {
						// add the sort parameter to the GET parameters array
						getParams.push(settings.sortVar);
						$(settings.sortUpdater).die('click').live('click',function(){
							// get the query parameter objects from the link location
							var linkParams = $.deparam.querystring($(this).attr('href'));
							// create a params object, settings the sortVar param to
							// null by default, and overriding it if it is in the query parameters
							var params = {};
							params[settings.sortVar] = null;
							if (linkParams[settings.sortVar] !== undefined) {
								params[settings.sortVar] = linkParams[settings.sortVar];
							}
							// update the current hash with the new pageVar state
							$.fn.yiiListView.updateHash(params);
							return false;
						});
					}
				} else {
					/* --------- */
					$(settings.updateSelector).die('click').live('click',function(){
						$.fn.yiiListView.update(id, {url: $(this).attr('href')});
						return false;
					});
					/* ********* */
				}
				/* --------- */
			}
			/* ********* */
			// if hash history is enabled and there are GET parameters to be monitored...
			if (settings.hashHistory && getParams !== undefined && getParams.length > 0) {
				var x = 0;
				var group = id + 'ListView';
				// register the hash handler group for this listview, using the first GET parameter
				$.registerHashHandler({
					parameter: getParams[x]
				}, getParams[x], group, {
					// the callback for the hash handler that allows us to update the listview
					// according to the current hash state
					callback: function(params, error) {
						// merge the current location and hash fragment parameters
						var pageParams = $.extend($.fn.yiiListView.parseParams($.param.querystring()), $.fn.yiiListView.parseParams($.param.fragment()));
						// loop through the changed parameters
						for(param in params) {
							if (params[param].next === null) {
								if (pageParams[param] !== undefined) {
									delete pageParams[param];
								}
							} else {
								// otherwise set it to the new value
								pageParams[param] = params[param].next;
							}
						}
						/*
						 * make the Ajax update, using the new Url, and overriding
						 * the error function to include the hash error callback
						 * (so that the hash can be changed back to its previous
						 * state in case of error)
						 */
						var data = '';
						// serialize the parameters
						for(param in pageParams) {
							data += '&' + param + '=' + pageParams[param];
						}
						// make the ajax update request
						$.fn.yiiListView.update(id, {
							url: window.location.href.split('?')[0],
							data: data,
							error: function(XMLHttpRequest, textStatus, errorThrown) {
								$('#'+id).removeClass(settings.loadingClass);
								alert(XMLHttpRequest.responseText);
								error();
							},
							cache: false
						});
					}
				});
				// loop through the remaining GET parameters, adding each one to the hash handler
				var y = getParams.length;
				for(x = 1; x < y; x++) {
					$.registerHashHandler({
						parameter: getParams[x]
					}, getParams[x], group, {});
				}
			}
			// trigger the hash change event for the first time in case an ajax link has been accessed
			$(window).trigger('hashchange');
			/* --------- */
		});
	};
	
	/* ********* */
	/**
	 * Updates the hash state with the parameter objects in params
	 * @param params the object containing the parameter objects to be updated
	 */
	$.fn.yiiListView.updateHash = function(params) {
		// get the parameters from the location
		var pageParams = $.deparam.querystring();
		// get the current hash state
		var state = $.bbq.getState();
		// merge the params with the current hash state
		state = $.extend(true, state, params);
		// check the updated params for null values or redundant values that can be left out of the hash state
		$.fn.yiiListView.checkCompare(params, pageParams, state);
		// update the hash state
		$.bbq.pushState(state, 2);
	};

	/**
	 * Parses a querystring to return an array of parameters
	 * @param querystring the querystring to be parsed
	 * @return object an associative array of parameter names and values
	 */
	$.fn.yiiListView.parseParams = function (querystring) {
	    var queryString = {};
	    decodeURI(querystring).replace(
	        new RegExp("([^?=&]+)(=([^&]*))?", "g"),
	        function($0, $1, $2, $3) { queryString[$1] = $3; }
	    );
	    return queryString;
	};
	
	/**
	 * Recursively sets the lowest-level child attribute of obj to null
	 * Used for objects generated from an Html form input tag, which
	 * means that each child attribute may only have one sub-attribute.
	 * @param obj the object whose lowest-level child is to be set to null
	 */
	$.fn.yiiListView.setNull = function(obj) {
		for (attr in obj) {
			if (typeof(obj[attr]) === 'object') {
				$.fn.yiiListView.setNull(obj[attr]);
			}
			else {
				obj[attr] = null;
			}
		}
	};
	
	/**
	 * Recursively compares a new params object, the parameters included in
	 * the current location, and the current hash state, to determine whether
	 * any parameters should be deleted from the hash state (such as those
	 * whose values are null, or whose values are the same as those given by
	 * the location Url)
	 * @param contained the params object containing the parameter objects
	 *                  that have been updated
	 * @param container the location parameters object containing the
	 *                  parameter objects derived from the location Url
	 * @param state the current state object containing the parameter objects
	 *              that may be updated.
	 */
	$.fn.yiiListView.checkCompare = function (contained, container, state) {
		for (attr in contained) {
			if (typeof(contained[attr]) === 'object') {
				var subContainer = null;
				if (container !== null && container[attr] !== undefined) {
					subContainer = container[attr];
				}
				if ($.fn.yiiListView.checkCompare(contained[attr], subContainer, state[attr])) {
					delete state[attr];
				}
			}
			else {
				if ((contained[attr] === null && (container === null || container[attr] === undefined)) || (container !== null && container[attr] !== undefined && contained[attr] === container[attr])) {
					delete state[attr];
				}
			}
		}
		var count = 0;
		for (attr in state) {
			count++;
		}
		if (count === 0) {
			return true;
		}
		return false;
	};

	$.fn.yiiListView.defaults = {
		ajaxUpdate: [],
		ajaxVar: 'ajax',
		pagerClass: 'pager',
		loadingClass: 'loading',
		sorterClass: 'sorter',
		hashHistory: false
		// updateSelector: '#id .pager a, '#id .sort a',
		// beforeAjaxUpdate: function(id) {},
		// afterAjaxUpdate: function(id, data) {},
	};

	/* reinsert the $.fn.yiiGridView objects that we backup up before
	   overwriting the constructor function */
	for(attr in tmpListView) {
		// only reinsert the object if we haven't created our own
		if ($.fn.yiiListView[attr] === undefined) {
			$.fn.yiiListView[attr] = tmpListView[attr];
		}
	}
}(jQuery));